__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from jina.executors.encoders.numeric import TransformEncoder

class RandomSparseEncoder(TransformEncoder):
    """
    :class:`RandomSparseEncoder` encodes data from an ndarray in size `B x T` into an ndarray in size `B x D`
    https://scikit-learn.org/stable/modules/generated/sklearn.random_projection.SparseRandomProjection.html
    """

    def post_init(self):
        super().post_init()
        if not self.model:
            from sklearn.random_projection import SparseRandomProjection
            self.model = SparseRandomProjection(n_components=self.output_dim, random_state=self.random_state)



"""This modules defines all kinds of exceptions raised in Jina."""

__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"


class NoExplicitMessage(Exception):
    """Waiting until all partial messages are received."""


class ChainedPodException(Exception):
    """Chained exception from the last Pod."""


class MismatchedVersion(SystemError):
    """When the jina version info of the incoming message does not match the local Jina version."""


class ExecutorFailToLoad(SystemError):
    """When the executor can not be loaded in pea/pod."""


class RuntimeFailToStart(SystemError):
    """When pea/pod is failed to started."""


class MemoryOverHighWatermark(Exception):
    """When the memory usage is over the defined high water mark."""


class DriverError(Exception):
    """Driver related exceptions."""


class RuntimeTerminated(KeyboardInterrupt):
    """The event loop of BasePea ends."""


class PodRunTimeError(Exception):
    """The error propagated by Pods when Executor throws an exception."""


class DriverNotInstalled(DriverError):
    """Driver is not installed in the BasePea."""


class NoDriverForRequest(DriverError):
    """No matched driver for this request."""


class UnattachedDriver(DriverError):
    """Driver is not attached to any BasePea or executor."""


class UnknownControlCommand(RuntimeError):
    """The control command received can not be recognized."""


class FlowTopologyError(Exception):
    """Flow exception when the topology is ambiguous."""


class FlowConnectivityError(Exception):
    """Flow exception when the flow is not connective via network."""


class FlowMissingPodError(Exception):
    """Flow exception when a pod can not be found in the flow."""


class FlowBuildLevelError(Exception):
    """Flow exception when required build level is higher than the current build level."""


class EmptyExecutorYAML(Exception):
    """The yaml config file is empty, nothing to read from there."""


class BadConfigSource(FileNotFoundError):
    """The yaml config file is bad, not loadable or not exist."""


class BadWorkspace(Exception):
    """Can not determine the separate storage strategy for the executor."""


class BadClient(Exception):
    """A wrongly defined grpc client, can not communicate with jina server correctly."""


class BadClientCallback(BadClient):
    """Error in the callback function on the client side."""


class BadClientInput(BadClient):
    """Error in the request generator function on the client side."""


class BadPersistantFile(Exception):
    """Bad or broken dump file that can not be deserialized with ``pickle.load``."""


class GRPCServerError(Exception):
    """Can not connect to the grpc gateway."""


class GatewayPartialMessage(Exception):
    """Gateway receives a multi-part message but it can not handle it."""


class UndefinedModel(Exception):
    """Any time a non-defined model is tried to be used."""


class MongoDBException(Exception):
    """Any errors raised by MongoDb."""


class TimedOutException(Exception):
    """Errors raised for timeout operations."""


class DockerLoginFailed(Exception):
    """Exception to raise for docker hub login failures."""


class ModelCheckpointNotExist(FileNotFoundError):
    """Exception to raise for executors depending on pretrained model files when they do not exist."""


class PretrainedModelFileDoesNotExist(ModelCheckpointNotExist):
    """Depreciated, used in the hub executors.

    TODO: to be removed after hub executors uses ModelCheckpointNotExist
    """


class HubBuilderError(Exception):
    """Base exception to raise when :command:`jina hub build` fails."""


class HubBuilderBuildError(HubBuilderError):
    """Exception to raise when :command:`jina hub build` fails to build image."""


class HubBuilderTestError(HubBuilderError):
    """Exception to raise when :command:`jina hub build` fails to test image."""


class CompressionRateTooLow(Exception):
    """Compression rate is too low, no need to compression."""


class DryRunException(Exception):
    """Dryrun is not successful on the given flow."""


class BadDocID(Exception):
    """Exception when user give a non-hex string as the doc id."""


class BadDocType(TypeError):
    """Exception when can not construct a document from the given data."""


class BadQueryLangType(TypeError):
    """Exception when can not construct a query language from the given data."""


class BadRequestType(TypeError):
    """Exception when can not construct a request object from given data."""


class BadNamedScoreType(TypeError):
    """Exception when can not construct a named score from the given data."""


class LengthMismatchException(Exception):
    """Exception when length of two items should be identical while not."""


class ImageAlreadyExists(Exception):
    """Exception when an image with the name, module version, and Jina version already exists on the Hub."""


class BadFlowYAMLVersion(Exception):
    """Exception when Flow YAML config specifies a wrong version number."""


class LookupyError(Exception):
    """Base exception class for all exceptions raised by lookupy."""


class EventLoopError(Exception):
    """Exception when a running event loop is found but not under jupyter or ipython."""


class ZMQSocketError(Exception):
    """Exception when ZMQlet/ZMQStreamlet can not be initialized."""


class HubLoginRequired(Exception):
    """Exception to raise for jina hub login."""


class DaemonConnectivityError(Exception):
    """Exception to raise when jina daemon is not connectable."""


class NotSupportedError(Exception):
    """Exeception when user accidentally using a retired argument."""



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import os

from . import BaseEncoder
from ..devices import OnnxDevice, PaddleDevice, TorchDevice, TFDevice, MindsporeDevice
from ...excepts import ModelCheckpointNotExist
from ...helper import is_url, cached_property


# mixin classes go first, base classes are read from right to left.
class BaseOnnxEncoder(OnnxDevice, BaseEncoder):
    """
    :class:`BasePaddleEncoder` is the base class for implementing Encoders with models from :mod:`onnxruntime` library.

    :param output_feature: the name of the layer for feature extraction.
    :param model_path: the path of the model in the format of `.onnx`. Check a list of available pretrained
        models at https://github.com/onnx/models#image_classification and download the git LFS to your local path.
        The ``model_path`` is the local path of the ``.onnx`` file, e.g. ``/tmp/onnx/mobilenetv2-1.0.onnx``.
    """
    def __init__(self, output_feature: str = None, model_path: str = None, *args, **kwargs):
        """Constructor."""
        super().__init__(*args, **kwargs)
        self.outputs_name = output_feature
        self.raw_model_path = model_path

    def post_init(self):
        """
        Load the model from the `.onnx` file and add outputs for the selected layer, i.e. ``outputs_name``. The modified
             models is saved at `tmp_model_path`.
        """
        super().post_init()
        model_name = self.raw_model_path.split('/')[-1] if self.raw_model_path else None
        tmp_model_path = self.get_file_from_workspace(f'{model_name}.tmp') if model_name else None
        raw_model_path = self.raw_model_path
        if self.raw_model_path and is_url(self.raw_model_path):
            import urllib.request
            download_path, *_ = urllib.request.urlretrieve(self.raw_model_path)
            raw_model_path = download_path
            self.logger.info(f'download the model at {self.raw_model_path}')
        if tmp_model_path and not os.path.exists(tmp_model_path) and self.outputs_name:
            self._append_outputs(raw_model_path, self.outputs_name, tmp_model_path)
            self.logger.info(f'save the model with outputs [{self.outputs_name}] at {tmp_model_path}')

        if tmp_model_path and os.path.exists(tmp_model_path):
            import onnxruntime
            self.model = onnxruntime.InferenceSession(tmp_model_path, None)
            self.inputs_name = self.model.get_inputs()[0].name
            self._device = None
            self.to_device(self.model)
        else:
            raise ModelCheckpointNotExist(f'model at {tmp_model_path} does not exist')

    @staticmethod
    def _append_outputs(input_fn, outputs_name_to_append, output_fn):
        import onnx
        model = onnx.load(input_fn)
        feature_map = onnx.helper.ValueInfoProto()
        feature_map.name = outputs_name_to_append
        model.graph.output.append(feature_map)
        onnx.save(model, output_fn)


class BaseTFEncoder(TFDevice, BaseEncoder):
    """:class:`BasePaddleEncoder` is the base class for implementing Encoders with models from :mod:`tensorflow` library."""
    pass


class BaseTorchEncoder(TorchDevice, BaseEncoder):
    """Base encoder class for :mod:`pytorch` library."""
    pass


class BasePaddleEncoder(PaddleDevice, BaseEncoder):
    """:class:`BasePaddleEncoder` is the base class for implementing Encoders with models from :mod:`paddlepaddle` library."""
    pass


class BaseMindsporeEncoder(MindsporeDevice, BaseEncoder):
    """
    :class:`BaseMindsporeEncoder` is the base class for implementing Encoders with models from `mindspore`.

    To implement your own executor with the :mod:`mindspore` lilbrary,

    .. highlight:: python
    .. code-block:: python
        import mindspore.nn as nn

        class YourAwesomeModel(nn.Cell):
            def __init__(self):
                ...

            def construct(self, x):
                ...

        class YourAwesomeEncoder(BaseMindsporeEncoder):
            def encode(self, data, *args, **kwargs):
                from mindspore import Tensor
                return self.model(Tensor(data)).asnumpy()

            def get_cell(self):
                return YourAwesomeModel()

    """

    def __init__(self, model_path: str = None, *args, **kwargs):
        """

        :param model_path: the path of the model's checkpoint.
        """
        super().__init__(*args, **kwargs)
        self.model_path = model_path

    def post_init(self):
        """
        Load the model from the `.ckpt` checkpoint.
        """
        super().post_init()
        if self.model_path and os.path.exists(self.model_path):
            self.to_device()
            from mindspore.train.serialization import load_checkpoint, load_param_into_net
            _param_dict = load_checkpoint(ckpt_file_name=self.model_path)
            load_param_into_net(self.model, _param_dict)
        else:
            raise ModelCheckpointNotExist(f'model {self.model_path} does not exist')

    @cached_property
    def model(self):
        """Get the Mindspore Neural Networks Cells."""
        return self.get_cell()

    def get_cell(self):
        """
        Return Mindspore Neural Networks Cells.

        Pre-defined building blocks or computing units to construct Neural Networks.
        A ``Cell`` could be a single neural network cell, such as conv2d, relu, batch_norm, etc.
        or a composition of cells to constructing a network.

        :return: :class:`mindspore.nn.Cell`
        """
        raise NotImplementedError



"""Module wrapping interactions with the remote Jina Hub."""
import base64
import json
from typing import Dict, Sequence, Any, Optional, List, Tuple
from urllib.error import HTTPError
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from pkg_resources import resource_stream

from .local import _fetch_access_token, _make_hub_table, _make_hub_table_with_local, _load_local_hub_manifest
from ...helper import colored
from ...importer import ImportExtensions
from ...jaml import JAML
from ...logging.profile import TimeContext


def _list(logger, image_name: str = None, image_kind: str = None,
          image_type: str = None, image_keywords: Sequence = ()) -> Optional[List[Dict[str, Any]]]:
    """Use Hub api to get the list of filtered images.

    :param logger: logger to use
    :param image_name: name of hub image
    :param image_kind: kind of hub image (indexer/encoder/segmenter/crafter/evaluator/ranker etc)
    :param image_type: type of hub image (pod/app)
    :param image_keywords: keywords added in the manifest yml
    :return: a dict of manifest specifications, each coresponds to a hub image
    """
    with resource_stream('jina', '/'.join(('resources', 'hubapi.yml'))) as fp:
        hubapi_yml = JAML.load(fp)
        hubapi_url = hubapi_yml['hubapi']['url'] + hubapi_yml['hubapi']['list']

    params = {
        'name': image_name,
        'kind': image_kind,
        'type': image_type,
        'keywords': image_keywords
    }
    params = {k: v for k, v in params.items() if v}
    if params:
        data = urlencode(params, doseq=True)
        request = Request(f'{hubapi_url}?{data}')
        with TimeContext('searching', logger):
            try:
                with urlopen(request) as resp:
                    response = json.load(resp)
            except HTTPError as err:
                if err.code == 400:
                    logger.warning('no matched executors found. please use different filters and retry.')
                elif err.code == 500:
                    logger.error(f'server is down: {err.reason}')
                else:
                    logger.error(f'unknown error: {err.reason}')
                return

        local_manifest = _load_local_hub_manifest()
        if local_manifest:
            tb = _make_hub_table_with_local(response, local_manifest)
        else:
            tb = _make_hub_table(response)
        logger.info('\n'.join(tb))
        return response


def _fetch_docker_auth(logger) -> Tuple[str, str]:
    """Use Hub api to get docker credentials.

    :param logger: the logger instance
    :return: a dict of specifying username and password
    """
    with resource_stream('jina', '/'.join(('resources', 'hubapi.yml'))) as fp:
        hubapi_yml = JAML.load(fp)
        hubapi_url = hubapi_yml['hubapi']['url'] + hubapi_yml['hubapi']['docker_auth']

    try:
        with ImportExtensions(required=True,
                              help_text='missing "requests" dependency, please do pip install "jina[http]"'):
            import requests
        headers = {
            'Accept': 'application/json',
            'authorizationToken': _fetch_access_token(logger)
        }
        response = requests.get(url=f'{hubapi_url}', headers=headers)
        if response.status_code != requests.codes.ok:
            logger.error(f'failed to fetch docker credentials. status code {response.status_code}')
        json_response = json.loads(response.text)
        username = base64.b64decode(json_response['docker_username']).decode('ascii')
        password = base64.b64decode(json_response['docker_password']).decode('ascii')
        logger.debug(f'Successfully fetched docker creds for user')
        return username, password
    except Exception as exp:
        logger.error(f'got an exception while fetching docker credentials {exp!r}')


def _register_to_mongodb(logger, summary: Optional[Dict] = None):
    """Hub API Invocation to run `hub push`.

    :param logger: the logger instance
    :param summary: the summary dict object
    """
    logger.info('registering image to Jina Hub database...')

    with resource_stream('jina', '/'.join(('resources', 'hubapi.yml'))) as fp:
        hubapi_yml = JAML.load(fp)
        hubapi_url = hubapi_yml['hubapi']['url'] + hubapi_yml['hubapi']['push']

    try:
        with ImportExtensions(required=True,
                              help_text='missing "requests" dependency, please do pip install "jina[http]"'):
            import requests
            headers = {
                'Accept': 'application/json',
                'authorizationToken': _fetch_access_token(logger)
            }
            response = requests.post(url=f'{hubapi_url}',
                                     headers=headers,
                                     data=json.dumps(summary))
            if response.status_code == requests.codes.ok:
                logger.info(response.text)
            elif response.status_code == requests.codes.unauthorized:
                logger.critical(f'user is unauthorized to perform push operation. '
                                f'please login using command: {colored("jina hub login", attrs=["bold"])}')
            elif response.status_code == requests.codes.internal_server_error:
                logger.critical(f'got an error from the API: {response.text}. If there are any authentication issues, '
                                f'please remember to login using command: '
                                f'{colored("jina hub login", attrs=["bold"])}')
    except Exception as exp:
        logger.error(f'got an exception while invoking hubapi for push {exp!r}')



from typing import Tuple

import numpy as np
from jina.executors.indexers.vector import BaseNumpyIndexer

if False:
    import SPTAG
class SptagIndexer(BaseNumpyIndexer):
    """
    :class:`SptagIndexer` SPTAG powered vector indexer.

    For SPTAG installation and python API usage, please consult:

        - https://github.com/microsoft/SPTAG/blob/master/Dockerfile
        - https://github.com/microsoft/SPTAG/blob/master/docs/Tutorial.ipynb
        - https://github.com/microsoft/SPTAG
        - https://github.com/microsoft/SPTAG/blob/master/docs/Parameters.md

    .. note::
        sptag package dependency is only required at the query time.
    """

    def __init__(self,
                 method: str = 'BKT',
                 samples: int = 1000,
                 tpt_number: int = 32,
                 tpt_leaf_size: int = 2000,
                 neighborhood_size: int = 32,
                 graph_neighborhood_size: int = 2,
                 cef: int = 100,
                 max_check_for_refined_graph: int = 10000,
                 num_threads: int = 1,
                 dist_calc_method: str = 'Cosine',
                 max_check: int = 8192,
                 bkt_number: int = 1,
                 bkt_meansk: int = 32,
                 kdt_number: int = 1,
                 *args, **kwargs):
        """
        Initialize an SptagIndexer
        :param method: The index method to use, index Algorithm type (e.g. BKT, KDT), required.
        :param samples: how many points will be sampled to do tree node split
        :param tpt_number: number of TPT trees to help with graph construction
        :param tpt_leaf_size: TPT tree leaf size
        :param neighborhood_size: number of neighbors each node has in the neighborhood graph
        :param graph_neighborhood_size: number of neighborhood size scale in the build stage
        :param cef: number of results used to construct RNG
        :param max_check_for_refined_graph: how many nodes each node will visit during graph refine in the build stage
        :param num_threads: The number of threads to use
        :param max_check: how many nodes will be visited for a query in the search stage
        :param dist_calc_method: the distance type, currently SPTAG only support Cosine and L2 distances.
        :param bkt_number: number of BKT trees (only used if method is BKT)
        :param bkt_meansk: how many childs each tree node has (only used if method is BKT)
        :param kdt_number: number of KDT trees (only used if method is BKT)
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.kdt_number = kdt_number
        self.bkt_meansk = bkt_meansk
        self.bkt_number = bkt_number
        self.max_check = max_check
        self.max_check_for_refined_graph = max_check_for_refined_graph
        self.cef = cef
        self.graph_neighborhood_size = graph_neighborhood_size
        self.neighborhood_size = neighborhood_size
        self.tpt_leaf_size = tpt_leaf_size
        self.tpt_number = tpt_number
        self.samples = samples
        self.method = method
        self.space = dist_calc_method
        self.num_threads = num_threads

    def build_advanced_index(self, vecs: 'np.ndarray') -> 'SPTAG.AnnIndex':
        """Build an advanced index structure from a numpy array.

        :param vecs: numpy array containing the vectors to index
        :return: advanced index
        """
        import SPTAG

        _index = SPTAG.AnnIndex(self.method, 'Float', vecs.shape[1])

        # Set the parameters
        _index.SetBuildParam("Samples", str(self.samples))
        _index.SetBuildParam("TPTNumber", str(self.tpt_number))
        _index.SetBuildParam("TPTLeafSize", str(self.tpt_leaf_size))
        _index.SetBuildParam("NeighborhoodSize", str(self.neighborhood_size))
        _index.SetBuildParam("GraphNeighborhoodScale", str(self.graph_neighborhood_size))
        _index.SetBuildParam("CEF", str(self.cef))
        _index.SetBuildParam("MaxCheckForRefineGraph", str(self.max_check_for_refined_graph))
        _index.SetBuildParam("NumberOfThreads", str(self.num_threads))
        _index.SetBuildParam("DistCalcMethod", self.space)
        _index.SetSearchParam("MaxCheck", str(self.max_check))
        _index.SetBuildParam("BKTNumber", str(self.bkt_number))
        _index.SetBuildParam("BKTMeansK", str(self.bkt_meansk))
        _index.SetBuildParam("KDTNumber", str(self.kdt_number))

        if _index.Build(vecs.astype('float32'), vecs.shape[0]):
            return _index

    def query(self, keys: 'np.ndarray', top_k: int, *args, **kwargs) -> Tuple['np.ndarray', 'np.ndarray']:
        """Find the top-k vectors with smallest ``metric`` and return their ids in ascending order.

        :param keys: numpy array containing vectors to search for
        :param top_k: upper limit of responses for each search vector
        :return: document ids and scores
        """
        idx = (np.ones((keys.shape[0], top_k)) * (-1)).astype(str)
        dist = np.ones((keys.shape[0], top_k)) * (-1)
        for r_id, k in enumerate(keys):
            _idx, _dist, _ = self.query_handler.Search(k, top_k)
            idx[r_id, :] = self._int2ext_id[np.array(_idx)]
            dist[r_id, :] = np.array(_dist)
        return idx, dist



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import warnings
from typing import Union, Tuple, Dict, Iterable

import numpy as np
from jina.executors.crafters import BaseCrafter

from .helper import _load_image, _move_channel_axis, _resize_short


class ImageResizer(BaseCrafter):
    """
    Resize image to the given size.

    :param target_size: Desired output size.
        If size is a sequence like (h, w), the output size will
        be matched to this. If size is an int, the smaller edge
        of the image will be matched to this number maintain
        the aspect ratio.
    :param how: The interpolation method. Valid values include
        `NEAREST`, `BILINEAR`, `BICUBIC`, and `LANCZOS`.
        Default is `BILINEAR`. Please refer to `PIL.Image` for details.
    """

    def __init__(self,
                 target_size: Union[Iterable[int], int] = 224,
                 how: str = 'BILINEAR',
                 channel_axis: int = -1,
                 *args, **kwargs):
        """Set Constructor."""
        warnings.warn(f'{self!r} will be retired soon, you can add `- !URI2Blob {{}}` to the driver', DeprecationWarning)
        super().__init__(*args, **kwargs)
        if isinstance(target_size, int):
            self.output_dim = target_size
        elif isinstance(target_size, Iterable):
            self.output_dim = tuple(target_size)
        else:
            raise ValueError(f'output_dim {target_size} should be an integer or tuple/list of 2 integers')
        self.how = how
        self.channel_axis = channel_axis

    def craft(self, blob: 'np.ndarray', *args, **kwargs) -> Dict:
        """
        Resize the image array to the given size.

        :param blob: The ndarray of the image
        :return: A dict with the cropped image
        """
        raw_img = _load_image(blob, self.channel_axis)
        _img = _resize_short(raw_img, self.output_dim, self.how)
        img = _move_channel_axis(np.asarray(_img), -1, self.channel_axis)
        return dict(offset=0, blob=img.astype('float32'))



from typing import List

from tensorflow import SparseTensor
from . import BaseSparseNdArray

if False:
    import numpy as np

__all__ = ['SparseNdArray']


class SparseNdArray(BaseSparseNdArray):
    """
    Tensorflow powered sparse ndarray, i.e. SparseTensor.

    .. seealso::
        https://www.tensorflow.org/api_docs/python/tf/sparse/SparseTensor
    """

    def sparse_constructor(self, indices: 'np.ndarray', values: 'np.ndarray', shape: List[int]) -> 'SparseTensor':
        """
        Sparse NdArray constructor for Tensorflow.

        :param indices: the indices of the sparse array
        :param values: the values of the sparse array
        :param shape: the shape of the sparse array
        :return: SparseTensor
        """
        return SparseTensor(indices, values, shape)

    def sparse_parser(self, value: 'SparseTensor'):
        """
        Parse a SparseTensor to indices, values and shape.

        :param value: the SparseTensor.
        :return: a Dict with three entries {'indices': ..., 'values':..., 'shape':...}
        """
        return {'indices': value.indices.numpy(),
                'values': value.values.numpy(),
                'shape': value.shape.as_list()}



"""Module containing the Base Client for Jina."""
__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import argparse
import os
from typing import Callable, Union, Optional, Iterator, List, Dict, AsyncIterator

import grpc
import inspect
from .helper import callback_exec
from .request import GeneratorSourceType
from ..enums import RequestType
from ..excepts import BadClient, BadClientInput
from ..helper import typename
from ..logging import default_logger, JinaLogger
from ..logging.profile import TimeContext, ProgressBar
from ..proto import jina_pb2_grpc
from ..types.request import Request

InputFnType = Union[GeneratorSourceType, Callable[..., GeneratorSourceType]]
CallbackFnType = Optional[Callable[..., None]]


class BaseClient:
    """A base client for connecting to the Flow Gateway.

    .. note::
        :class:`BaseClient` does not provide `train`, `index`, `search` interfaces.
        Please use :class:`Client` or :class:`AsyncClient`.

    :param args: the Namespace from argparse
    """

    def __init__(self, args: 'argparse.Namespace'):
        self.args = args
        self.logger = JinaLogger(self.__class__.__name__, **vars(args))

        if not args.proxy and os.name != 'nt':
            # (Han 2020 12.12): gRPC channel is over HTTP2 and it does not work when we have proxy
            # as many enterprise users are behind proxy, a quick way to
            # surpass it is by temporally unset proxy. Please do NOT panic as it will NOT
            # affect users os-level envs.
            os.unsetenv('http_proxy')
            os.unsetenv('https_proxy')
        self._mode = args.mode
        self._input_fn = None

    @property
    def mode(self) -> str:
        """
        Get the mode for this client (index, query etc.).

        :return: Mode of the client.
        """
        return self._mode

    @mode.setter
    def mode(self, value: RequestType) -> None:
        """
        Set the mode.

        :param value: Request type. (e.g. INDEX, SEARCH, DELETE, UPDATE, CONTROL, TRAIN)
        """
        if isinstance(value, RequestType):
            self._mode = value
            self.args.mode = value
        else:
            raise ValueError(f'{value} must be one of {RequestType}')

    @staticmethod
    def check_input(input_fn: Optional[InputFnType] = None, **kwargs) -> None:
        """Validate the input_fn and print the first request if success.

        :param input_fn: the input function
        :param kwargs: keyword arguments
        """
        if hasattr(input_fn, '__call__'):
            input_fn = input_fn()

        kwargs['data'] = input_fn

        if inspect.isasyncgenfunction(input_fn) or inspect.isasyncgen(input_fn):
            raise NotImplementedError('checking the validity of an async generator is not implemented yet')

        try:
            from .request import request_generator
            r = next(request_generator(**kwargs))
            if isinstance(r, Request):
                default_logger.success(f'input_fn is valid')
            else:
                raise TypeError(f'{typename(r)} is not a valid Request')
        except Exception as ex:
            default_logger.error(f'input_fn is not valid!')
            raise BadClientInput from ex

    def _get_requests(self, **kwargs) -> Union[Iterator['Request'], AsyncIterator['Request']]:
        """
        Get request in generator.

        :param kwargs: Keyword arguments.
        :return: Iterator of request.
        """
        _kwargs = vars(self.args)
        _kwargs['data'] = self.input_fn
        # override by the caller-specific kwargs
        _kwargs.update(kwargs)

        if inspect.isasyncgen(self.input_fn):
            from .request.asyncio import request_generator
            return request_generator(**_kwargs)
        else:
            from .request import request_generator
            return request_generator(**_kwargs)

    def _get_task_name(self, kwargs: Dict) -> str:
        tname = str(self.mode).lower()
        if 'mode' in kwargs:
            tname = str(kwargs['mode']).lower()
        return tname

    @property
    def input_fn(self) -> InputFnType:
        """
        An iterator of bytes, each element represents a Document's raw content.

        ``input_fn`` defined in the protobuf

        :return: input function
        """
        if self._input_fn is not None:
            return self._input_fn
        else:
            raise BadClient('input_fn is empty or not set')

    @input_fn.setter
    def input_fn(self, bytes_gen: InputFnType) -> None:
        """
        Set the input data.

        :param bytes_gen: input function type
        """
        if hasattr(bytes_gen, '__call__'):
            self._input_fn = bytes_gen()
        else:
            self._input_fn = bytes_gen

    async def _get_results(self,
                           input_fn: Callable,
                           on_done: Callable,
                           on_error: Callable = None,
                           on_always: Callable = None, **kwargs):
        try:
            self.input_fn = input_fn
            tname = self._get_task_name(kwargs)
            req_iter = self._get_requests(**kwargs)
            async with grpc.aio.insecure_channel(f'{self.args.host}:{self.args.port_expose}',
                                                 options=[('grpc.max_send_message_length', -1),
                                                          ('grpc.max_receive_message_length', -1)]) as channel:
                stub = jina_pb2_grpc.JinaRPCStub(channel)
                self.logger.success(f'connected to the gateway at {self.args.host}:{self.args.port_expose}!')
                with ProgressBar(task_name=tname) as p_bar, TimeContext(tname):
                    async for resp in stub.Call(req_iter):
                        resp.as_typed_request(resp.request_type)
                        resp.as_response()
                        callback_exec(response=resp,
                                      on_error=on_error,
                                      on_done=on_done,
                                      on_always=on_always,
                                      continue_on_error=self.args.continue_on_error,
                                      logger=self.logger)
                        p_bar.update(self.args.request_size)
                        yield resp
        except KeyboardInterrupt:
            self.logger.warning('user cancel the process')
        except grpc.aio._call.AioRpcError as rpc_ex:
            # Since this object is guaranteed to be a grpc.Call, might as well include that in its name.
            my_code = rpc_ex.code()
            my_details = rpc_ex.details()
            msg = f'gRPC error: {my_code} {my_details}'
            if my_code == grpc.StatusCode.UNAVAILABLE:
                self.logger.error(
                    f'{msg}\nthe ongoing request is terminated as the server is not available or closed already')
                raise rpc_ex
            elif my_code == grpc.StatusCode.INTERNAL:
                self.logger.error(f'{msg}\ninternal error on the server side')
                raise rpc_ex
            elif my_code == grpc.StatusCode.UNKNOWN and 'asyncio.exceptions.TimeoutError' in my_details:
                raise BadClientInput(f'{msg}\n'
                                     'often the case is that you define/send a bad input iterator to jina, '
                                     'please double check your input iterator') from rpc_ex
            else:
                raise BadClient(msg) from rpc_ex

    def index(self):
        """Issue 'index' request to the Flow."""
        raise NotImplementedError

    def search(self):
        """Issue 'search' request to the Flow."""
        raise NotImplementedError

    def train(self):
        """Issue 'train' request to the Flow."""
        raise NotImplementedError

    @staticmethod
    def add_default_kwargs(kwargs: Dict):
        """
        Add the default kwargs to the instance.

        :param kwargs: the kwargs to add
        """
        # TODO: refactor it into load from config file
        if ('top_k' in kwargs) and (kwargs['top_k'] is not None):
            # associate all VectorSearchDriver and SliceQL driver to use top_k
            from jina import QueryLang
            topk_ql = [QueryLang({'name': 'SliceQL', 'priority': 1, 'parameters': {'end': kwargs['top_k']}}),
                       QueryLang(
                           {'name': 'VectorSearchDriver', 'priority': 1, 'parameters': {'top_k': kwargs['top_k']}})]
            if 'queryset' not in kwargs:
                kwargs['queryset'] = topk_ql
            else:
                kwargs['queryset'].extend(topk_ql)



import mimetypes
import struct
import urllib.parse
import urllib.request
import zlib

import numpy as np


def _png_to_buffer_1d(arr: 'np.ndarray', width: int, height: int) -> bytes:
    pixels = []
    arr = 255 - arr
    for p in arr[::-1]:
        pixels.extend([p, p, p, 255])
    buf = bytearray(pixels)

    # reverse the vertical line order and add null bytes at the start
    width_byte_4 = width * 4
    raw_data = b''.join(
        b'\x00' + buf[span:span + width_byte_4]
        for span in range((height - 1) * width_byte_4, -1, - width_byte_4))

    def png_pack(png_tag, data):
        chunk_head = png_tag + data
        return (struct.pack('!I', len(data))
                + chunk_head
                + struct.pack('!I', 0xFFFFFFFF & zlib.crc32(chunk_head)))

    png_bytes = b''.join([
        b'\x89PNG\r\n\x1a\n',
        png_pack(b'IHDR', struct.pack('!2I5B', width, height, 8, 6, 0, 0, 0)),
        png_pack(b'IDAT', zlib.compress(raw_data, 9)),
        png_pack(b'IEND', b'')])

    return png_bytes


def _pillow_image_to_buffer(image, image_format: str) -> bytes:
    import io
    img_byte_arr = io.BytesIO()
    image.save(img_byte_arr, format=image_format)
    img_byte_arr = img_byte_arr.getvalue()
    return img_byte_arr


def png_to_buffer(arr: 'np.ndarray', width: int, height: int, resize_method: str):
    """
    Convert png to buffer bytes.

    :param arr: Data representations of the png.
    :param width: Width of the png.
    :param height: Height of the png.
    :param resize_method: Resize methods (e.g. `NEAREST`, `BILINEAR`, `BICUBIC`, and `LANCZOS`).
    :return: Png in buffer bytes.
    """
    arr = arr.astype(np.uint8)

    if arr.ndim == 1:
        png_bytes = _png_to_buffer_1d(arr, width, height)
    elif arr.ndim == 2:
        from PIL import Image
        im = Image.fromarray(arr).convert('L')
        im = im.resize((width, height), getattr(Image, resize_method))
        png_bytes = _pillow_image_to_buffer(im, image_format='PNG')
    elif arr.ndim == 3:
        from PIL import Image
        im = Image.fromarray(arr).convert('RGB')
        im = im.resize((width, height), getattr(Image, resize_method))
        png_bytes = _pillow_image_to_buffer(im, image_format='PNG')
    else:
        raise ValueError(f'ndim={len(arr.shape)} array is not supported')

    return png_bytes


def to_image_blob(source, color_axis: int = -1) -> 'np.ndarray':
    """
    Convert an image buffer to blob

    :param source: image bytes buffer
    :param color_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
    :return: image blob
    """
    from PIL import Image
    raw_img = Image.open(source).convert('RGB')
    img = np.array(raw_img).astype('float32')
    if color_axis != -1:
        img = np.moveaxis(img, -1, color_axis)
    return img


def to_datauri(mimetype, data, charset: str = 'utf-8', base64: bool = False, binary: bool = True):
    """
    Convert data to data URI.

    :param mimetype: MIME types (e.g. 'text/plain','image/png' etc.)
    :param data: Data representations.
    :param charset: Charset may be any character set registered with IANA
    :param base64: Used to encode arbitrary octet sequences into a form that satisfies the rules of 7bit. Designed to be efficient for non-text 8 bit and binary data. Sometimes used for text data that frequently uses non-US-ASCII characters.
    :param binary: True if from binary data False for other data (e.g. text)
    :return: URI data
    """
    parts = ['data:', mimetype]
    if charset is not None:
        parts.extend([';charset=', charset])
    if base64:
        parts.append(';base64')
        from base64 import encodebytes as encode64
        if binary:
            encoded_data = encode64(data).decode(charset).replace('\n', '').strip()
        else:
            encoded_data = encode64(data).strip()
    else:
        from urllib.parse import quote_from_bytes, quote
        if binary:
            encoded_data = quote_from_bytes(data)
        else:
            encoded_data = quote(data)
    parts.extend([',', encoded_data])
    return ''.join(parts)


def guess_mime(uri):
    """
    Guess the type of a file based on its URL.

    :param uri: Data URI.
    :return: MIME type.
    """
    # guess when uri points to a local file
    m_type = mimetypes.guess_type(uri)[0]
    # guess when uri points to a remote file
    if not m_type and urllib.parse.urlparse(uri).scheme in {'http', 'https', 'data'}:
        page = urllib.request.Request(uri, headers={'User-Agent': 'Mozilla/5.0'})
        tmp = urllib.request.urlopen(page)
        m_type = tmp.info().get_content_type()
    return m_type
